;; Generated package description from auctex-label-numbers.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "auctex-label-numbers" "0.3" "Numbering for LaTeX previews and folds" '((emacs "27.1") (auctex "14.0.5")) :commit "b399826caeb854073a5d3435133f5a77a7f9f683" :authors '(("Paul D. Nelson" . "nelson.paul.david@gmail.com")) :maintainer '("Paul D. Nelson" . "nelson.paul.david@gmail.com") :keywords '("tex") :url "https://github.com/ultronozm/auctex-label-numbers.el")
